;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;
;;; sql-controller.lisp -- Interface to a CLSQL based object store.
;;; 
;;; Initial version 10/12/2005 by Robert L. Read
;;; <read@robertlread.net>
;;; 
;;; part of
;;;
;;; Elephant: an object-oriented database for Common Lisp
;;;
;;; Copyright (c) 2005 by Robert L. Read
;;;
;;; This program is released under the following license
;;; ("GPL").  For differenct licensing terms, contact the
;;; copyright holders.
;;;
;;; This program is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU General
;;; Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at
;;; your option) any later version.
;;;
;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE. See the GNU General Public License
;;; for more details.
;;;
;;; The GNU General Public License can be found in the file
;;; LICENSE which should have been distributed with this
;;; code.  It can also be found at
;;;
;;; http://www.opensource.org/licenses/gpl-license.php
;;;
;;; You should have received a copy of the GNU General
;;; Public License along with this program; if not, write
;;; to the Free Software Foundation, Inc., 59 Temple Place,
;;; Suite 330, Boston, MA 02111-1307 USA
;;;

(in-package "ELEPHANT")

;;; other clsql packages would have to be added for 
;;; non-postgresql databases, see the CL-SQL documentation
(eval-when ( :compile-toplevel :load-toplevel)
  (asdf:oos 'asdf:load-op :clsql)

;; Probably must be customized ... see documentation on installin postgres.
  (defvar *clsql-foreign-lib-path* "/usr/lib")
  (clsql:push-library-path *clsql-foreign-lib-path*)
  (clsql:push-library-path *elephant-lib-path*)

;;  (asdf:oos 'asdf:load-op :clsql-postgresql-socket)
  )

(defmacro with-transaction-sql ((&key 
				 (store-controller-sql '*store-controller*))
				 &body body)
  "Execute a body with a transaction in place.  On success,
the transaction is committed.  Otherwise, the transaction is
aborted.  If the body deadlocks, the body is re-executed in
a new transaction, retrying a fixed number of iterations.
*auto-commit* is false for the body of the transaction."
  `(if (typep ,store-controller-sql 'elephant::sql-store-controller)
       (if (clsql::in-transaction-p 
 	    :database 
 	    (controller-db ,store-controller-sql))
 	   (progn
 	     ,@body)
 	   (prog2
 	       (clsql::set-autocommit nil)
 	       (clsql::with-transaction
 		   (:database 
 		    (controller-db ,store-controller-sql))
 		 ,@body)
 	     (clsql::set-autocommit t)
 	     ))))

(defclass sql-store-controller (store-controller)
  ((dbonnection-spec :type list :accessor :dbcn-spc :initarg :dbconnection-spec
		     ;; for postgres, this is host, db, user, password
		     ;; If you can't get the lisp system to connect with 
		     ;; this default information, make sure you can connect 
		     ;; to the database called "test" under the user postgress 
		     ;; with the psql console first.  Then study the authorization
		     ;; and configuration files.
		     :initform '("localhost.localdomain" "test" "postgres" "")
		     )
   )
  (:documentation  "Class of objects responsible for the
book-keeping of holding DB handles, the cache, table
creation, counters, locks, the root (for garbage collection,)
et cetera.  This is the Postgresql-specific subclass of store-controller.")
  )

(defmethod build-btree ((sc sql-store-controller))
  (make-sql-btree sc)
  )

(defmethod get-transaction-macro-symbol ((sc sql-store-controller))
  'with-transaction-sql
  )


(defun sql-test-and-construct (spec)
  (if (sql-store-spec-p spec)
      (open-store-sql spec)
      nil)
  )

(eval-when ( :load-toplevel)
  (register-strategy 'sql-test-and-construct)
  )

(defmacro with-open-store-sql ((spec) &body body)
  "Executes the body with an open controller,
unconditionally closing the controller on exit."
  `(let ((*store-controller* 
	  (make-instance 'sql-store-controller :dbconnection-spec ,spec)))
     (declare (special *store-controller*))
     (open-controller *store-controller*)
     (unwind-protect
	  (progn ,@body)
       (close-controller *store-controller*))))

(defun open-store-sql (spec  &key (recover nil)
		       (recover-fatal nil) (thread t))
  "Conveniently open a store controller."
  (setq *store-controller*  
	(if (sql-store-spec-p spec)
	    (make-instance 'sql-store-controller :dbconnection-spec spec)
	    (error (format nil "uninterpretable path/spec specifier: ~A" spec)))
	)
  (open-controller *store-controller* :recover recover 
		   :recover-fatal recover-fatal :thread thread)
  )

;; When you build one of these, you have to put in the connection spec.
(defclass sql-btree (btree) 
  (
   )
  (:documentation "A SQL implementation of a BTree"))

(defmethod get-value (key (bt sql-btree))
  (let* ((sc (check-con (:dbcn-spc-pst bt)))
	 (con (controller-db sc))
	 (v (sql-get-from-clcn (oid bt) key sc con)))
	   (if v 
	       (values v t)
	       (values nil nil))))
	 

(defmethod existsp (key (bt sql-btree))
  (let* ((sc (check-con (:dbcn-spc-pst bt)))
	 (con (controller-db sc)))
    (sql-from-clcn-existsp (oid bt) key  con)
    )
  )

(defmethod (setf get-value) (value key (bt sql-btree))
  (let* ((sc (check-con (:dbcn-spc-pst bt)))
	 (con (controller-db sc)))
    (sql-add-to-clcn (oid bt) key value sc con)
    )
  )
(defmethod remove-kv (key (bt sql-btree))
  (let* ((sc (check-con (:dbcn-spc-pst bt)))
	 (con (controller-db sc)))
    (sql-remove-one-from-clcn (oid bt)
			      key
			      sc
			      con))
  )


;; Because these things are transient, I can't move them
;; directly into the class above.  I am not sure how best to
;; handle this problem.
(defclass sql-indexed-btree (indexed-btree sql-btree )
  (
   (indices :accessor indices :initform (make-hash-table)
	    )
   (indices-cache :accessor indices-cache :initform (make-hash-table)
		  :transient t)
   )
  (:metaclass persistent-metaclass)
  (:documentation "A SQL-based BTree that supports secondary indices."))

(defmethod build-indexed-btree ((sc sql-store-controller))
  (let ((bt (make-instance 'sql-indexed-btree :sc sc)))
    (setf (:dbcn-spc-pst bt) (:dbcn-spc sc))
    bt
    ))

(defmethod build-btree-index ((sc sql-store-controller) &key primary key-form)
  (let ((bt (make-instance 'sql-btree-index :primary primary :key-form key-form :sc sc)))
    (setf (:dbcn-spc-pst bt) (:dbcn-spc sc))
    bt
    ))


;; I need some way to get to the store-controller here...
;; I could be the store controller in the hash table, that's probably
;; the simplest thing to do..
(defmethod add-index ((bt sql-indexed-btree) &key index-name key-form populate)
  (let* ((sc (check-con (:dbcn-spc-pst bt)))
	(con (controller-db sc)))
    (if (and (not (null index-name))
	     (symbolp index-name) (or (symbolp key-form) (listp key-form)))
	(let ((indices (indices bt))
	      (index (make-instance 'sql-btree-index :primary bt 
				    :key-form key-form
				    :sc sc)))
	  (setf (gethash index-name (indices-cache bt)) index)
	  (setf (gethash index-name indices) index)
	  (setf (indices bt) indices)
	  (when populate
	    (let ((key-fn (key-fn index))
		  )
	      (with-transaction-sql (:store-controller-sql sc)
		(map-btree
		 #'(lambda (k v)
		     (multiple-value-bind (index? secondary-key)
			 (funcall key-fn index k v)
;; This is a slow, DB cycle intensive operation.  It could chunked somehow,
;; I think, probably making it 10 times faster.
		       (when index?
			 (sql-add-to-clcn (oid index)
					  secondary-key
					  k
					  sc con :insert-only t)
			 )))
		 bt))))
	  index)
	(error "Invalid index initargs!"))))



(defmethod (setf get-value) (value key (bt sql-indexed-btree))
  "Set a key / value pair, and update secondary indices."
  (let* ((sc (check-con (:dbcn-spc-pst bt)))
	 (con (controller-db sc))
	 (indices (indices-cache bt)))
    (with-transaction-sql (:store-controller-sql sc)
      (maphash 
       #'(lambda (k index) 
	   (multiple-value-bind (index? secondary-key)
	       (funcall (key-fn index) index key value)
	     (when index?
	       (sql-add-to-clcn (oid index)
				secondary-key
				key
				sc con :insert-only t)
	       )))
       indices)
      ;; Now we place the actual value
      (sql-add-to-clcn (oid bt) key value sc con)
      )
    value))

(defmethod remove-kv (key (bt sql-indexed-btree))
  "Remove a key / value pair, and update secondary indices."
  (declare (optimize (speed 3)))
  (let* (
	 (sc (check-con (:dbcn-spc-pst bt)))
	 (con (controller-db sc)))
    (with-transaction-sql (:store-controller-sql sc)
      (let ((value (get-value key bt)))
	(when value
	  (let ((indices (indices-cache bt)))
	    (maphash 
	     #'(lambda (k index) 
		 (multiple-value-bind (index? secondary-key)
		     (funcall (key-fn index) index key value)
		   (when index?
		     ;; This function will in fact remove all of the 
		     ;; duplicate keys; but this is not how the BDB system works.
		     ;; It appears to me, based on the behavior of tests, that 
		     ;; this should remove the FIRST row that match not all.
		     (sql-remove-key-and-value-from-clcn (oid index)
							 secondary-key
							 key
							 con)
		     ;; And furthermore, we have to remove the index entry
		     ;;		     (remove-kv secondary-key index)
		     )))
	     indices)
	    ;; Now we place the actual value
	    (sql-remove-from-clcn (oid bt) key sc con))
	  )
	value))))



(defclass sql-btree-index (btree-index sql-btree)
  ()
  (:metaclass persistent-metaclass)
  (:documentation "A SQL-based BTree supports secondary indices."))


(clsql::locally-enable-sql-reader-syntax) 

;; Check that the table exists and is in proper form.
;; If it is not in proper form, signal an error, no 
;; way to recover from that automatically.  If it 
;; does not exist, return nil so we can create it later!

;; These functions are probably not cross-database portable...
(defun keyvalue-table-exists (con)
  ;; we want to use ":owner :all" because we don't really care who created
  ;; the table, as long as we have the rights we need!
  (clsql:table-exists-p [keyvalue] :database con :owner :all)
  )

;; This is just an initial version; it is possible that 
;; we might someday wish to use blobs instead; certainly, I am
;; storing blobs now in the Berkeley-db and we meed to make sure 
;; we are properly testing that.  However, blobs are awkward to 
;; handle, so I am going to do this first...
(defun create-keyvalue-table (con)
  ;; the "serial" specifiation here does not seem to work, (
  ;; apparently not supported by clsql, so I have to execute these
  ;; commands specifically.  This may be a database-dependent way of doing
  ;; things, but sequences in general are NOT standardized across RDBMS.
  ;; I prefer sequence to support the "get-next-oid" command, but there 
  ;; ARE other ways of doing it that could make this more portable.
  ;;    (execute-command create :database con)
  ;;    (execute-command idx-id :database con)
  ;;    (execute-command idx-key :database con)
  ;; Danger:  Rather than use 'serial as a type, CLSQL appears to support
  ;; CREATE-SEQUENCE and SEQUENCE-NEXT.  That would solve our problem!

  ;; ALL OF THIS needs to be inside a transaction.
  (clsql::create-table [keyvalue]
		'(
		  ([clctn_id] integer :not-null)
		  ([key] text :not-null)
		  ([value] text)
		  ) :database con
		)
  ;; 	      :constraints '("PRIMARY KEY (clctn_id key)"
  ;;				     "UNIQUE (clctn_id,key)")

  ;; apparently in postgres this is failing pretty awfully because 
  ;; sequence-exists-p return nil and then we get an error that the sequence exists!
  ;;    (unless (sequence-exists-p [persistent_seq])
  (clsql::create-sequence [persistent_seq]
		   :database con)
  ;;)
  ;;    (unless (index-exists-p [idx_clctn_id])
  (clsql::create-index [idx_clctn_id] :on [keyvalue]
		:attributes '([clctn_id])
		:database con)
  ;; )
  ;;    (unless (index-exists-p [idx_key])
  (clsql::create-index [idx_key] :on [keyvalue]
		:attributes '([key])
		:database con)
  ;;)
  ;; This is actually unique
  ;;    (unless (index-exists-p [idx_both])
  (clsql:create-index [idx_both] :on [keyvalue]
		:attributes '([clctn_id] [key])
		:database con)
  ;;)
  )


(defmethod open-controller ((sc sql-store-controller)
			    ;; At present these three have no meaning
			    &key 
			    (recover nil)
			    (recover-fatal nil) 
			    (thread t))
  (the sql-store-controller


    

    (let ((con (clsql:connect (:dbcn-spc sc)
;; WARNING: This line of code forces us to use postgresql.
;; If this were parametrized upwards we could concievably try 
;; other backends.
			      :database-type :postgresql
			      :if-exists :old)))
      (setf (gethash (:dbcn-spc sc) *dbconnection-spec*) sc)
      (setf (slot-value sc 'db) con)
      ;; Now we should make sure that the KEYVALUE table exists, and, if 
      ;; it does not, we need to create it..
      ;; This kind of thing is typically database-specific, but at least we 
      ;; can put it in a function....
      (unless (keyvalue-table-exists con)
	(create-keyvalue-table con))
      (setf (slot-value sc 'root) (make-sql-btree sc))
      sc)
    )
  )

(defun make-sql-btree (sc)
  (let ((bt (make-instance 'sql-btree :sc sc)))
    (setf (:dbcn-spc-pst bt) (:dbcn-spc sc))
    bt)
  )

(defmethod close-controller ((sc sql-store-controller))
  (when (slot-value sc 'db)
    ;; close the conneciton
    ;; (actually clsql has pooling and other complications, I am not sure
    ;; that this is complete.)
    (clsql:disconnect :database (controller-db sc))
    (setf (slot-value sc 'root) nil)
    ))


;; Because this is part of the public
;; interface that I'm tied to, it has to accept a store-controller...
(defmethod next-oid ((sc sql-store-controller ))
  (let ((con (controller-db sc)))
    (clsql:sequence-next [persistent_seq]
		   :database con))
  )


;; if add-to-root is a method, then we can make it class dependent...
;; otherwise we have to change the original code.  There is 
;; almost no way to implement this without either changing the existing
;; file.  If we can introduce a layer of class indirectio there, then
;; we can control things properly.  In the meantime, I will implement
;; a proper method myself, but I will give it a name so it doesn't 
;; conflict with 'add-to-root.  'add-to-root can remain a convenience symbol,
;; that will end up calling this routine!
(defmethod sql-add-to-root (key value (pgsc sql-store-controller ) con)
  (sql-add-to-clcn 0 key value pgsc con)
  )
;;(defmethod sql-add-to-root (key value dbcon)
;;  (sql-add-to-clcn 0 key value sc dbcon)
;;  )

(defmethod sql-add-to-clcn ((clcn integer) key value sc con
			    &key (insert-only nil))
  (let (
	(vbs 
	 (serialize-to-base64-string value))
	(kbs
	 (serialize-to-base64-string key))
	)
    (if (and (not insert-only) (sql-get-from-clcn clcn key sc con))
	(clsql::update-records [keyvalue]
			:av-pairs `((key ,kbs)
				    (clctn_id ,clcn)
				    (value ,vbs))
			:where [and [= [clctn_id] clcn] [= [key] kbs]]
			:database con)
	(clsql::insert-records :into [keyvalue]
			:attributes '(key clctn_id value)
			:values (list kbs clcn vbs)
			:database con
			))
    )
  value
  )



(defmethod sql-get-from-root (key sc con)
  (let ((v (sql-get-from-clcn 0 key sc con)))
    (if v 
	(values v t)
	(values nil nil)))
  )

;; This is a major difference betwen SQL and BDB:
;; BDB plans to give you one value and let you iterate, but
;; SQL by nature returns a set of values (when the keys aren't unique.)
;; 
;; I serious problem here is what to do if the things aren't unique.
;; According to the Elepahnt documentation, you should get one value 
;; (not clear which one, the "first" probably, and then use a 
;; cursor to iterate over duplicates.  
;; So although it is moderately clear how the cursor is supposed to 
;; work, I'm not sure how I'm supposed to know what value should be 
;; returend by this non-cursor function.
;; I suspect if I return the value that has the lowest OID, that will
;; match the behavior of the sleepycat function....
;; To do that I have to read in all of the values and deserialized them
;; This could be a good reason to keep the oids out, and separte, in 
;; a separate column.
(defmethod sql-get-from-clcn ((clcn integer) key sc con)
  (sql-get-from-clcn-nth clcn key sc con 0)
  )
(defmethod sql-get-from-clcn-nth ((clcn integer) key sc con (n integer))
  (let* (
	 (kbs 
	  (serialize-to-base64-string key))
	 (tuples
	  (clsql::select [value] 
		  :from [keyvalue]
		  :where [and [= [clctn_id] clcn] [= [key] kbs]]
		  :database con
		  )))
    ;; Get the lowest value by sorting and taking the first value;
    ;; this isn't a very good way to do things...
    ;; Note also that this will be extremely inefficient if 
    ;; you have for example, a boolean index function.
    ;; I could parametrize this routine to take an "nth"
    ;; parameter.  But there is almost no way to implement
    ;; that efficiently without changing the database structure;
    ;; but that's OK, I could add a column to support that 
    ;; relatively easily later on.
    (if (< (length tuples) n)
	nil
	(nth n (sort 
		(mapcar 
		 #'(lambda (x)
		     (deserialize-from-base64-string (car x) :sc sc))
		 tuples)
		#'my-generic-less-than)))))

(defmethod sql-get-from-clcn-cnt ((clcn integer) key con)
  (let* (
	 (kbs (serialize-to-base64-string key))
	 (tuples
	  (clsql::select [count [value]]
		  :from [keyvalue]
		  :where [and [= [clctn_id] clcn] [= [key] kbs]]
		  :database con
		  )))
    (caar tuples)))

(defmethod sql-dump-clcn ((clcn integer) sc con)
  (let* (
	 (tuples
	  (clsql::select [key] [value]
		  :from [keyvalue]
		  :where [and [= [clctn_id] clcn]]
		  :database con
		  )))
    (mapcar #'(lambda (x) (mapcar #'(lambda (q) (deserialize-from-base64-string :sc sc)) x))
	    tuples)))

(defmethod sql-from-root-existsp (key con)
  (sql-get-from-clcn 0 key con)
  )

(defmethod sql-from-clcn-existsp ((clcn integer) key con)
  (let* (
	 (kbs (with-buffer-streams (out-buf)
		(serialize-to-base64-string key))
	   )
	 (tuples
	  (clsql::select [value] 
		  :from [keyvalue]
		  :where [and [= [clctn_id] clcn] [= [key] kbs]]
		  :database con
		  )))
    (if tuples
	t
	nil)
    ))

(defmethod sql-remove-from-root (key sc con)
  (sql-remove-from-clcn 0 key sc con)
  )

(defmethod sql-remove-from-clcn ((clcn integer) key sc con)
  (let (
	(kbs (serialize-to-base64-string key))
	)
    (clsql::delete-records :from [keyvalue]
		    :where [and [= [clctn_id] clcn] [= [key] kbs]]
		    :database con
		    ))
  )
(defmethod sql-remove-one-from-clcn ((clcn integer) key sc con)
  (let* (
	 (kbs (serialize-to-base64-string key))
	;; We want to remove the FIRST value, based on our ordering.
	;; have little choice but to read everything in and delete based on
	;; the "value field".
	(tuples
	 (clsql::select [value] 
		 :from [keyvalue]
		 :where [and [= [clctn_id] clcn] [= [key] kbs]]
		 :database con
		 )))
    (if (< (length tuples) 1)
	nil
	(let ((to-remove
	       (serialize-to-base64-string 
		(nth 0 (sort 
			(mapcar 
			 #'(lambda (x)
			     (deserialize-from-base64-string (car x) :sc sc))
			 tuples)
			#'my-generic-less-than)))))
	  (clsql::delete-records :from [keyvalue]
			  :where [and [= [clctn_id] clcn] [= [key] kbs]
			  [= [value] to-remove]]
			  :database con
			  )
	  )
	)
    ))

(defmethod sql-remove-key-and-value-from-clcn ((clcn integer) key value con)
  (let* (
	 (kbs (serialize-to-base64-string key))
	 (vbs (serialize-to-base64-string value)))
	  (clsql::delete-records :from [keyvalue]
			  :where [and [= [clctn_id] clcn] [= [key] kbs]
			  [= [value] vbs]]
			  :database con
			  )
	  ))
	
(clsql::restore-sql-reader-syntax-state) 




(defmethod persistent-slot-writer-aux ((sc sql-store-controller) new-value instance name)
  (let* ((con (controller-db sc)))
	 (sql-add-to-root
	  (form-slot-key (oid instance) name)
	  new-value
	  sc con)
	 ))

;; This was almost ncecessary to allow this functionality to be included
;; only if you load ele-clsql.  It could also be used in bdb, and probably
;; should be, but there is some strange macro stuff there that I am afraid 
;; to change, so I am implementing it only here.
(defmethod persistent-slot-reader-aux ((sc sql-store-controller) instance name)
  (let* ((con (controller-db sc)))
    (let ((v 
	   (sql-get-from-root
	    (form-slot-key (oid instance) name)
	    sc con
	    )))
      (if v
	  v
	  (error  'unbound-slot :instance instance :name name))))
  )

(defmethod persistent-slot-boundp-aux ((sc sql-store-controller) instance name)
  (let* ((con (controller-db sc)))
	   (if (sql-get-from-root
		(form-slot-key (oid instance) name)
		sc con )
	       t nil)))



